#!/usr/bin/env python3

import os
import sys
import argparse
import subprocess
from pathlib import Path

from ImmunoViewer.tile import DeepZoomStaticTiler

def find_files(path, min_depth, max_depth):
    for root, dirs, files in os.walk(str(path)):
        depth = root.count(os.path.sep) - str(path).count(os.path.sep)
        if min_depth <= depth <= max_depth:
            for file in files:
                if file.endswith('.tif'):
                    yield Path(root) / file

def main():
    parser = argparse.ArgumentParser(description="Process TIF files in a directory")
    parser.add_argument('data_dir', help='The data directory containing the TIF files')
    parser.add_argument('-t', '--num_cores', type=int, default=4, help='Number of cores to use (default: 4)')
    args = parser.parse_args()

    script_dir = Path(__file__).parent.resolve()
    python_tile = script_dir / "src/ImmunoViewer/tile.py"

    if not args.data_dir:
        print("Usage: process_folder.py [-t num_cores] data_dir")
        sys.exit(1)

    num_cores = args.num_cores if args.num_cores else 4

    data_dir = Path(args.data_dir)
    all_files = list(find_files(data_dir, 1, 2))

    for file in all_files:
        print(f"Processing {file}")
        sample_name = file.stem
        tif_name = file.name
        folder = file.parent

        print(f"check if {folder / (sample_name + '_files')} exists")
        if not (folder / f"{sample_name}_files").is_dir():
            print("folder does not exist, running python script")

            basename = os.path.join(folder, os.path.splitext(os.path.basename(file))[0])

            print(f"running for {file} basename is {basename}")

            DeepZoomStaticTiler(
                file,
                basename,
                "jpeg",
                254,
                1,
                True,
                100,
                num_cores,
                False,
            ).run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
# deepzoom_tile - Convert whole-slide images to Deep Zoom format
#
# Copyright (c) 2010-2015 Carnegie Mellon University
# Copyright (c) 2022      Benjamin Gilbert
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""An example program to generate a Deep Zoom directory tree from a slide."""

import json
from multiprocessing import JoinableQueue, Process
from optparse import OptionParser
import os
import re
import shutil
import sys
from unicodedata import normalize

import openslide
from openslide import ImageSlide, open_slide
from openslide.deepzoom import DeepZoomGenerator

VIEWER_SLIDE_NAME = 'slide'


class TileWorker(Process):
    """A child process that generates and writes tiles."""

    def __init__(self, queue, slidepath, tile_size, overlap, limit_bounds, quality):
        Process.__init__(self, name='TileWorker')
        self.daemon = True
        self._queue = queue
        self._slidepath = slidepath
        self._tile_size = tile_size
        self._overlap = overlap
        self._limit_bounds = limit_bounds
        self._quality = quality
        self._slide = None

    def run(self):
        self._slide = open_slide(self._slidepath)
        last_associated = None
        dz = self._get_dz()
        while True:
            data = self._queue.get()
            if data is None:
                self._queue.task_done()
                break
            associated, level, address, outfile = data
            if last_associated != associated:
                dz = self._get_dz(associated)
                last_associated = associated
            tile = dz.get_tile(level, address)
            tile.save(outfile, quality=self._quality)
            self._queue.task_done()

    def _get_dz(self, associated=None):
        if associated is not None:
            image = ImageSlide(self._slide.associated_images[associated])
        else:
            image = self._slide
        return DeepZoomGenerator(
            image, self._tile_size, self._overlap, limit_bounds=self._limit_bounds
        )


class DeepZoomImageTiler:
    """Handles generation of tiles and metadata for a single image."""

    def __init__(self, dz, basename, format, associated, queue):
        self._dz = dz
        self._basename = basename
        self._format = format
        self._associated = associated
        self._queue = queue
        self._processed = 0

    def run(self):
        self._write_tiles()
        self._write_dzi()

    def _write_tiles(self):
        for level in range(self._dz.level_count):
            tiledir = os.path.join("%s_files" % self._basename, str(level))
            if not os.path.exists(tiledir):
                os.makedirs(tiledir)
            cols, rows = self._dz.level_tiles[level]
            for row in range(rows):
                for col in range(cols):
                    tilename = os.path.join(
                        tiledir, '%d_%d.%s' % (col, row, self._format)
                    )
                    if not os.path.exists(tilename):
                        self._queue.put((self._associated, level, (col, row), tilename))
                    self._tile_done()

    def _tile_done(self):
        self._processed += 1
        count, total = self._processed, self._dz.tile_count
        if count % 100 == 0 or count == total:
            print(
                "Tiling %s: wrote %d/%d tiles"
                % (self._associated or 'slide', count, total),
                end='\r',
                file=sys.stderr,
            )
            if count == total:
                print(file=sys.stderr)

    def _write_dzi(self):
        with open('%s.dzi' % self._basename, 'w') as fh:
            fh.write(self.get_dzi())

    def get_dzi(self):
        return self._dz.get_dzi(self._format)


class DeepZoomStaticTiler:
    """Handles generation of tiles and metadata for all images in a slide."""

    def __init__(
        self,
        slidepath,
        basename,
        format,
        tile_size,
        overlap,
        limit_bounds,
        quality,
        workers,
        with_viewer,
    ):
        if with_viewer:
            # Check extra dependency before doing a bunch of work
            import jinja2  # noqa: F401  module-imported-but-unused
        self._slide = open_slide(slidepath)
        self._basename = basename
        self._format = format
        self._tile_size = tile_size
        self._overlap = overlap
        self._limit_bounds = limit_bounds
        self._queue = JoinableQueue(2 * workers)
        self._workers = workers
        self._with_viewer = with_viewer
        self._dzi_data = {}
        for _i in range(workers):
            TileWorker(
                self._queue, slidepath, tile_size, overlap, limit_bounds, quality
            ).start()

    def run(self):
        self._run_image()
        if self._with_viewer:
            for name in self._slide.associated_images:
                self._run_image(name)
            self._write_html()
            self._write_static()
        self._shutdown()

    def _run_image(self, associated=None):
        """Run a single image from self._slide."""
        if associated is None:
            image = self._slide
            if self._with_viewer:
                basename = os.path.join(self._basename, VIEWER_SLIDE_NAME)
            else:
                basename = self._basename
        else:
            image = ImageSlide(self._slide.associated_images[associated])
            basename = os.path.join(self._basename, self._slugify(associated))
        dz = DeepZoomGenerator(
            image, self._tile_size, self._overlap, limit_bounds=self._limit_bounds
        )
        tiler = DeepZoomImageTiler(dz, basename, self._format, associated, self._queue)
        tiler.run()
        self._dzi_data[self._url_for(associated)] = tiler.get_dzi()

    def _url_for(self, associated):
        if associated is None:
            base = VIEWER_SLIDE_NAME
        else:
            base = self._slugify(associated)
        return '%s.dzi' % base

    def _write_html(self):
        import jinja2

        # https://docs.python.org/3/reference/import.html#main-spec
        if __spec__ is not None:
            # We're running from a module (e.g. "python -m deepzoom_tile")
            # so load templates from the containing package.
            loader = jinja2.PackageLoader('__main__')
        else:
            # We're not running from a module (e.g. "python deepzoom_tile.py")
            # so PackageLoader('__main__') doesn't work in jinja2 3.x.
            # Load templates directly from the filesystem.
            loader = jinja2.FileSystemLoader(
                os.path.join(os.path.dirname(__file__), 'templates')
            )
        env = jinja2.Environment(loader=loader, autoescape=True)
        template = env.get_template('slide-multipane.html')
        associated_urls = {n: self._url_for(n) for n in self._slide.associated_images}
        try:
            mpp_x = self._slide.properties[openslide.PROPERTY_NAME_MPP_X]
            mpp_y = self._slide.properties[openslide.PROPERTY_NAME_MPP_Y]
            mpp = (float(mpp_x) + float(mpp_y)) / 2
        except (KeyError, ValueError):
            mpp = 0
        # Embed the dzi metadata in the HTML to work around Chrome's
        # refusal to allow XmlHttpRequest from file:///, even when
        # the originating page is also a file:///
        data = template.render(
            slide_url=self._url_for(None),
            slide_mpp=mpp,
            associated=associated_urls,
            properties=self._slide.properties,
            dzi_data=json.dumps(self._dzi_data),
        )
        with open(os.path.join(self._basename, 'index.html'), 'w') as fh:
            fh.write(data)

    def _write_static(self):
        basesrc = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'static')
        basedst = os.path.join(self._basename, 'static')
        self._copydir(basesrc, basedst)
        self._copydir(os.path.join(basesrc, 'images'), os.path.join(basedst, 'images'))

    def _copydir(self, src, dest):
        if not os.path.exists(dest):
            os.makedirs(dest)
        for name in os.listdir(src):
            srcpath = os.path.join(src, name)
            if os.path.isfile(srcpath):
                shutil.copy(srcpath, os.path.join(dest, name))

    @classmethod
    def _slugify(cls, text):
        text = normalize('NFKD', text.lower()).encode('ascii', 'ignore').decode()
        return re.sub('[^a-z0-9]+', '_', text)

    def _shutdown(self):
        for _i in range(self._workers):
            self._queue.put(None)
        self._queue.join()


from io import BytesIO
from collections import OrderedDict
from optparse import OptionParser
import os
from threading import Lock
import re
from unicodedata import normalize
import json
import cv2
import matplotlib.pyplot as plt
import numpy as np

from flask import Flask, render_template, Response, request, send_from_directory, send_file
from flask_cors import CORS, cross_origin

app = Flask(__name__)

cors = CORS(app)
app.config['CORS_HEADERS'] = 'Content-Type'

def mix_channels(images, chs, gains):

    color_mapping = {
        'green': (0, 255, 0),
        'red': (0, 0, 255),
        'blue': (255, 0, 0),
        'yellow': (0, 255, 255),
        'cyan': (255, 255, 0),
        'white': (255, 255, 255),
        'black': (0, 0, 0),
    }

    merged_image = np.zeros((*images[0].shape, 3), dtype=np.uint8)

    for image, ch, gain in zip(images, chs, gains):
        color_rgb = color_mapping[ch]
        enhanced_image = cv2.convertScaleAbs(image, alpha=gain)
        colored_image = cv2.merge([enhanced_image * (color_rgb[i] // 255) for i in range(3)])
        merged_image = cv2.add(merged_image, colored_image)

    return merged_image

def find_directories_with_files_folders(base_dir):
    directories_with_files = []

    for root, dirs, files in os.walk(base_dir):
        for dirname in dirs:
            if dirname.endswith("_files"):
                buf = {}
                buf['name'] = os.path.relpath(root, base_dir)
                dirs.sort()
                buf['files'] = dirs
                directories_with_files.append(buf)

                buf['details'] = {}

                if os.path.isfile(f'{root}.json'):
                        with open(f'{root}.json', 'r') as f:
                            data = json.load(f)
                            buf['details'] = data
                break

    return directories_with_files



@app.route('/', defaults={'path': ''})
@app.route('/<path:path>')
@cross_origin()
def index(path):
    if path and (path.startswith('assets') or path.startswith('favicon') or path.startswith('images')):
        return send_from_directory(os.path.abspath("client/dist"), path)
    else:
        return send_from_directory(os.path.abspath("client/dist"), 'index.html')

@app.route('/samples.json')
@cross_origin()
def samples():
    file_json = find_directories_with_files_folders(os.path.abspath(app.config['SLIDE_DIR']))

    buf = {}
    buf['samples'] = file_json

    resp = Response(json.dumps(buf), status=200, mimetype='application/json')
    return resp

@app.route('/<string:files>/<string:chs>/<string:gains>/<path:file>.dzi')
@cross_origin()
def dzi(files, chs, gains, file):
    path_to_dzi = os.path.join(os.path.abspath(app.config['SLIDE_DIR']), file)

    for _, _, files in os.walk(path_to_dzi):
        for file in files:
            if file.endswith(".dzi"):
                main_ch = file

    resp = send_from_directory(path_to_dzi, main_ch)

    resp.mimetype = 'application/xml'
    return resp

@app.route('/<string:files>/<string:chs>/<string:gains>/<path:file>_files/<int:level>/<string:loc>')
@cross_origin()
def tile(files, chs, gains, file, level, loc):
    files = files.split(';') 
    chs = chs.split(';')
    gains = gains.split(';')
    gains = [int(x) for x in gains]

    if len(files) == 1:
        merged_image = cv2.imread(os.path.abspath(os.path.join(app.config['SLIDE_DIR'], file, f'{files[0]}', str(level), loc)))
    else:
        merge_images = []
        merge_chs = []
        merge_gains = []
        for i, ch in enumerate(chs):
            if ch != 'empty':
                buf = cv2.imread(os.path.abspath(os.path.join(app.config['SLIDE_DIR'], file, f'{files[i]}', str(level), loc)))
                buf = cv2.cvtColor(buf, cv2.COLOR_BGR2GRAY)
                merge_images.append(buf)
                merge_chs.append(ch)
                merge_gains.append(gains[i])

        if len(merge_images) == 0:
            merged_image = np.zeros((256, 256), dtype=np.uint8)
        else:
            merged_image = mix_channels(merge_images, merge_chs, merge_gains)

    img_bytes = cv2.imencode('.jpeg', merged_image)[1].tobytes()
    img_io = BytesIO(img_bytes)

    return send_file(img_io, mimetype='image/jpeg')

@app.route('/read/<path:file>')
@cross_origin()
def open_json(file):
    if os.path.isfile(os.path.abspath(os.path.join(app.config['SLIDE_DIR'], f'{file}.json'))):
        # Read the JSON file
        with open(os.path.abspath(os.path.join(app.config['SLIDE_DIR'], f'{file}.json')), 'r') as f:
            data = json.load(f)

    else: 
        data = {"notes": "", "name": "", "stains": ""}; 

    resp = Response(json.dumps(data), mimetype='application/json')
    return resp


@app.route('/save/<path:file>', methods=['GET', 'POST'])
@cross_origin()
def save_note(file):
    if request.method == 'POST':
        data = json.loads(request.data)

        with open(os.path.abspath(os.path.join(app.config['SLIDE_DIR'], f'{file}.json')), 'w') as f:
            json.dump(data, f)

    return "OK"

def main():
    parser = OptionParser(usage='Usage: %prog [options] [folder]')

    parser.add_option(
        '-l',
        '--listen',
        metavar='ADDRESS',
        dest='host',
        default='0.0.0.0',
        help='address to listen on [0.0.0.0]',
    )
    parser.add_option(
        '-p',
        '--port',
        metavar='PORT',
        dest='port',
        type='int',
        default=5000,
        help='port to listen on [5000]',
    )

    (opts, args) = parser.parse_args()

    try:
        app.config['SLIDE_DIR'] = args[0]
    except IndexError:
        if app.config['SLIDE_DIR'] is None:
            parser.error('No slide file specified')

    app.run(host=opts.host, port=opts.port, threaded=True, debug=True)



import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    version='0.1.0',
    name='ImmunoViewer',
    author='David van IJzendoorn',
    author_email='davidvanijzendoorn@gmail.com',
    description='Discover and annotate your (multi-channel) big TIF files with this user-friendly viewer',
    keywords='big tif, immuno, viewer, annotate, annotation, discover, discovery, image, images, tif, tiff, multi-channel, multi c',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/davidvi/ImmunoViewer',
    project_urls={
        'Documentation': 'https://github.com/davidvi/ImmunoViewer',
        'Bug Reports':
        'https://github.com/davidvi/ImmunoViewer/issues',
        'Source Code': 'https://github.com/davidvi/ImmunoViewer'
    },
    package_dir={'': 'src'},
    packages=setuptools.find_packages(where='src'),
    classifiers=[
        # see https://pypi.org/classifiers/
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3 :: Only',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=[
        'flask==2.2.3',
        'numpy==1.24.2',
        'openslide-python==1.2.0',
        'pillow==9.4.0',
        'opencv-python==4.7.0.72',
        'matplotlib==3.7.1',
        'Flask-Cors==3.0.10',
        ],
    entry_points={
    'console_scripts': [
        'ImmunoViewerProcess=ImmunoViewer.process_folder:main',
        'ImmunoViewerServer=ImmunoViewer.server:main',
    ],
},
)